;; Copyright (C) 2016N Antonin Houska
;;
;; This file is part of PGQA.
;;
;; PGQA is free software: you can redistribute it and/or modify it under the
;; terms of the GNU General Public License as published by the Free Software
;; Foundation, either version 3 of the License, or (at your option) any later
;; version.

;; PGQA is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
;; details.

;; You should have received a copy of the GNU General Public License qalong
;; with PGQA. If not, see <http://www.gnu.org/licenses/>.

(require 'pgqa-node)

(defclass pgqa-problem ()
  (
   (message
    :initarg :message
    "Message that user should see")

   (location
    :initarg :location
    "Integer (marker) position the problem was found at, or a vector containing both start and end position.")
   )

  "A problem identified during query analysis."
  )

(defclass pgqa-analyze-context ()
  (
   (problems
    :initarg :problems
    :initform nil
    :documentation "List of errors / warnings encountered during analysis."
    )
   )

  "Information passed across query tree nodes during analysis."
  )

(defun pgqa-show-problems (problems buffer)
  "Show problems in the appropriate buffer."

  ;; Make sure that the output buffer exists.
  (setq pgqa-log-buffer (get-buffer-create pgqa-log-buffer))

  ;; Display the errors / warnings.
  (with-current-buffer pgqa-log-buffer
    (atomic-change-group
      (setq buffer-read-only nil)
      (erase-buffer)
      (if (car problems)
	  (progn
	    (let ((button)
		  (pos)
		  (i 0))
	      (dolist (p problems)
		;; Separate the message from the previous one.
		(if (> i 0)
		    (insert "\n"))

		(setq pos (point))
		(insert (format "%s\n" (oref p message)))

		(when (null noninteractive)
		  ;; Create a text button so that user can navigate to the
		  ;; related node by clicking on the problem text.
		  (setq button
			(make-text-button pos (1- (point))))

		  ;; Set the location of the problem is available to the
		  ;; action function(s).
		  (button-put button 'err-loc (oref p location))
		  (button-put button 'query-buffer buffer)
		  ;; Mouse click should also lead to the problem location.
		  (button-put button 'follow-link t)

		  (button-put button 'action
			      (lambda (button)
				(let* ((loc (button-get button 'err-loc))
				       (buf (button-get button
							'query-buffer)))
				  ;; Point at the problem location.
				  (with-current-buffer buf
				    (switch-to-buffer-other-window
				     (current-buffer))

				    (if (sequencep loc)
					;; Both start and end position is
					;; known.
					(progn
					  (goto-char (elt loc 0))
					  ;; TODO Use save-mark-and-excursion
					  (push-mark)
					  ;; Mark the problem location.
					  (activate-mark)
					  (goto-char (elt loc 1))
					  ;; Wait a little bit before
					  ;; deactivating the mark.
					  (run-at-time ".5 sec" nil
						       (lambda ()
							 (goto-char (mark))
							 (pop-mark))))
				      ;; Only the start position is known.
				      (goto-char loc))
				    )
				  )
				)
			      )
		  )

		;; Get ready for the next message.
		(setq pos (point))
		(setq i (1+ i)))
	      )
	    ;; Let user see the buffer.
	    (switch-to-buffer-other-window (current-buffer)))
	(insert "No problems found"))
      ;; Mark the buffer read-only so that user does not mess it up when
      ;; hitting ENTER to navigate to the error locations.
      (setq buffer-read-only t))
    )

    (when (car problems)
      ;; Do not let the processing continue, for whatever reason the function
      ;; is called. This must happen outside atomic-change-group so that the
      ;; pgqa-log-buffer setup is not undone.
      ;;
      ;; Actually not all problems identified during the analysis break
      ;; query formatting, however if we proceeded, then the links to
      ;; the query text would become invalid.
      ;;
      ;; XXX In the batch mode, print out the contents of
      ;; pgqa-log-buffer to standard output.
      (user-error "%d problem(s) found in the query" (length problems)))
  )

(defmethod pgqa-analyze-node ((node pgqa-node) context)
  "Analyze query tree node and add error message / warning to the output list.

Note that new errors are added to the beginning of the list as it seems to be
cheaper. If the order matters, caller can simply call `reverse' on the result."

  ;; If particular node subclass does not have this method defined, no
  ;; analysis will take place.
  )

(defmethod pgqa-analyze-node ((node pgqa-from-list-entry) context)
  (let ((out-list (oref context problems)))
    ;; Parser should currently enforce the alias, but the error message is not
    ;; too user friendly. So we might allow absence of the alias in the
    ;; grammar someday and the analyzer complain here.
    (when (string= (oref node kind) "query")
      (if (null (slot-boundp node 'alias))
	  (push
	   (make-instance 'pgqa-problem
			  :message "Subquery must have an alizs"
			  :location node)
	   out-list))
      (let ((alias (oref node alias)))
	(when (slot-boundp alias 'cols)
	  (let* (
		 ;; The actual query is the first arg of the FROM list entry
		 ;; node.
		 (query (car (oref node args)))
		 (query-tlist (oref query target-expr))
		 (query-cols (oref query-tlist args))
		 (cols-tmp query-cols)
		 (has-asterisk))
	    ;; If the targetlist contanis an asterisk, we don't know how many
	    ;; output columns the query actually has.
	    (while (and (null has-asterisk) cols-tmp)
	      (let ((expr (oref (car cols-tmp) expr)))
		(cl-assert (eq (eieio-object-class expr) 'pgqa-obj)
			   "Target list entry has wrong data type.")
		(dolist (e (oref expr args))
		  (if (string= e "*")
		      (setq has-asterisk t))
		  )
		)
	      (setq cols-tmp (cdr cols-tmp)))

	    (unless has-asterisk
	      (let* ((ncols-query (length query-cols))
		     (arg-list (oref alias cols))
                     (ncols-alias (length (oref arg-list args))))
		(if (>
		     ncols-alias
		     ncols-query)
		    (push
		     (make-instance
		      'pgqa-problem
		      :message
		      (format
		       "Subquery has %d columns available but %d columns specified in the alias"
		       ncols-query ncols-alias)
		      :location (oref alias region))
		     out-list))
		)
	      )
	    )
	  )
	)
      (oset context problems out-list))
    )
  )

(provide 'pgqa-analyze)
