;; Copyright (C) 2016 Antonin Houska
;;
;; This file is part of PGQA.
;;
;; PGQA is free software: you can redistribute it and/or modify it under the
;; terms of the GNU General Public License as published by the Free Software
;; Foundation, either version 3 of the License, or (at your option) any later
;; version.

;; PGQA is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
;; details.

;; You should have received a copy of the GNU General Public License qalong
;; with PGQA. If not, see <http://www.gnu.org/licenses/>.

;; This module prints the query tree in machine readable format.

(eval-when-compile
  (require 'subr-x))

(require 'pgqa-node)

(defcustom pgqa-query-tree-print-region t
  "If non-nil, the output of `pgqa-dump-raw' contains node region info.")

;; Metadata to control printing out the internal representation of a query
;; tree node.
(defclass pgqa-dump-raw-state ()
  (
   (node-start
    :initarg :node-start
    :documentation
    "Function that returns the initial part of the node.

The funcion receives the node and indentation level as the first and second
arguments respectively.")

   (node-end
    :initarg :node-end
    :documentation
    "Function that returns the final part of the node.

The funcion receives the node and indentation level as the first and second
arguments respectively.")

   (result
    :initarg :result
    :documentation "String to which each node appends its textual
representation.")
   )
  )

(defun pgqa-dump-raw-append (state str)
  "Append STR to the current result."

  (oset state result (format "%s%s" (oref state result) str)))

;; Convenience macro to call `node-start' or `node-end' function of the
;; `state' and update the `result' accordingly.
(defmacro pgqa-dump-raw-node (kind node state indent)
  `(let ((func (oref state ,kind))
	 (orig (oref ,state result)))
     (oset ,state result
	   (string-join
	    (list
	     orig
	     (funcall func ,node ,indent))))))

(defun pgqa-node-to-lisp-start (node indent)
  "Turn the query tree node into the initial part of lisp-like expression."

  (let ((region "")
	(node-name)
	(indent-str (pgqa-lisp-indent-string indent)))
    (setq node-name (symbol-name (eieio-object-class-name node)))
    ;; The class name w/o the leading "pgqa-" part seems to be an useful node
    ;; name.
    (setq node-name (substring node-name (length "pgqa-") (length node-name)))

    (if pgqa-query-tree-print-region
	(setq region (format " %s" (oref node region))))
    (format "%s(%s%s\n" indent-str node-name region)))

(defun pgqa-node-to-lisp-end (node indent)
  "Turn the query tree node slot into the final part of lisp-like expression."

  (format "%s)\n" (pgqa-lisp-indent-string (1+ indent))))

;; Subroutine for the macros below.
(defun pgqa-slot-to-lisp-common (node state slot-name indent)
  (oset state result
	(format "%s%s:%s"
		(oref state result)
		(pgqa-lisp-indent-string indent)
		slot-name))
  )

(defmacro pgqa-simple-slot-to-lisp (node state slot-name indent)
  "Turn a simple slot of the query tree node into a string."

  `(progn
     (pgqa-slot-to-lisp-common
      ,node
      ,state
      (quote ,slot-name)
      ,indent)

     (pgqa-dump-raw-append ,state
			   (format " %s\n"
				   (oref ,node ,slot-name)))
     )
  )

;; Like above but the slot is a node itself.
(defmacro pgqa-node-slot-to-lisp (node state slot-name indent)
  "Turn a node slot of the query tree node into a string."

  `(progn
     (pgqa-slot-to-lisp-common
      ,node
      ,state
      (format "%s%s" (quote ,slot-name) "\n")
      ,indent)

     (pgqa-dump-raw (oref ,node ,slot-name) ,state (1+ ,indent)))
  )

;; Like above but the slot is a list.
(defmacro pgqa-list-slot-to-lisp (node state slot-name indent)
  "Turn a list slot of the query tree node into a string."

  `(progn
     (pgqa-slot-to-lisp-common
      ,node
      ,state
      (format "%s%s" (quote ,slot-name) "\n")
      ,indent)

     (pgqa-dump-raw-append ,state
			   (format "%s(\n"
				   (pgqa-lisp-indent-string (1+ ,indent))))

     (dolist (i (oref ,node ,slot-name))
       (pgqa-dump-raw i state (+ ,indent 2)))

     (pgqa-dump-raw-append ,state
			   (format "%s)\n"
				   (pgqa-lisp-indent-string (+ ,indent 2))))
     )
  )

;; Create string for given indentation level, for the lisp output.
(defun pgqa-lisp-indent-string (indent)
  (make-string (* indent 2) 32))

;; `state' is an instance of `pgqa-dump-raw-state' class.
;;
;; `indent' is the number of ancestor nodes.
(defmethod pgqa-dump-raw ((node pgqa-node) state indent)
  "Print out the internal representation of a query node."

  (pgqa-dump-raw-node node-start node state indent)
  (pgqa-dump-raw-node node-end node state indent))


(defmethod pgqa-dump-raw ((node pgqa-query) state indent)
  "Print out the internal representation of a query."

  (pgqa-dump-raw-node node-start node state indent)

  (pgqa-node-slot-to-lisp node state target-expr (1+ indent))
  (if (oref node from-expr)
      (pgqa-node-slot-to-lisp node state from-expr (1+ indent)))

  (if (oref node group-by)
      (pgqa-node-slot-to-lisp node state group-by (1+ indent)))
  (if (oref node having)
      (pgqa-node-slot-to-lisp node state having (1+ indent)))
  (if (oref node order-by)
      (pgqa-node-slot-to-lisp node state order-by (1+ indent)))
  (if (oref node target-table)
      (pgqa-node-slot-to-lisp node state target-table (1+ indent)))
  (if (oref node insert-cols)
      (pgqa-node-slot-to-lisp node state insert-cols (1+ indent)))

  (pgqa-dump-raw-node node-end node state indent))

(defmethod pgqa-dump-raw ((node pgqa-from-expr) state indent)
  (pgqa-dump-raw-node node-start node state indent)

  (if (slot-boundp node 'from-list)
      (pgqa-list-slot-to-lisp node state from-list (1+ indent)))

  (if (slot-boundp node 'qual)
      (pgqa-node-slot-to-lisp node state qual (1+ indent)))

  (pgqa-dump-raw-node node-end node state indent))

(defmethod pgqa-dump-raw ((node pgqa-from-list-entry) state indent)
  (pgqa-dump-raw-node node-start node state indent)

  (if (slot-boundp node 'alias)
      (pgqa-node-slot-to-lisp node state alias (1+ indent)))

  (if (slot-boundp node 'kind)
      (pgqa-simple-slot-to-lisp node state kind (1+ indent)))

  (if (slot-boundp node 'args)
      (pgqa-list-slot-to-lisp node state args (1+ indent)))

  (if (slot-boundp node 'qual)
      (pgqa-node-slot-to-lisp node state qual (1+ indent)))

  (pgqa-dump-raw-node node-end node state indent))

(defmethod pgqa-dump-raw ((node pgqa-from-list-entry-alias) state indent)
  (pgqa-dump-raw-node node-start node state indent)

  (pgqa-node-slot-to-lisp node state name (1+ indent))
  (if (slot-boundp node 'cols)
      (pgqa-node-slot-to-lisp node state cols (1+ indent)))

  (pgqa-dump-raw-node node-end node state indent))

(defmethod pgqa-dump-raw ((node pgqa-top-clause) state indent)
  (pgqa-dump-raw-node node-start node state indent)

  (if (slot-boundp node 'expr)
      (pgqa-node-slot-to-lisp node state expr (1+ indent)))

  (pgqa-dump-raw-node node-end node state indent))

(defmethod pgqa-dump-raw ((node pgqa-func-call) state indent)
  (pgqa-dump-raw-node node-start node state indent)

  (pgqa-node-slot-to-lisp node state name (1+ indent))

  (when (and (slot-boundp node 'args) (oref node args))
    (cl-assert
     (object-of-class-p (oref node args) 'pgqa-node-list)
     nil
     "Function arguments should have been wrapped in a comma  operator")

    (let ((comma-op (oref node args)))
      (pgqa-list-slot-to-lisp comma-op state args (1+ indent))))

  (pgqa-dump-raw-node node-end node state indent))

(defmethod pgqa-dump-raw ((node pgqa-string) state indent)
  (pgqa-dump-raw-node node-start node state indent)

  (pgqa-dump-raw-append state
			(format "%s%s\n"
				(pgqa-lisp-indent-string (1+ indent))
				(oref node value)))

  (pgqa-dump-raw-node node-end node state indent))

(defmethod pgqa-dump-raw ((node pgqa-number) state indent)
  (pgqa-dump-raw-node node-start node state indent)

  (pgqa-dump-raw-append state
			(format "%s%s\n"
				(pgqa-lisp-indent-string (1+ indent))
				(oref node value)))

  (pgqa-dump-raw-node node-end node state indent))

(defmethod pgqa-dump-raw ((node pgqa-obj) state indent)
  (pgqa-dump-raw-node node-start node state indent)

  (pgqa-dump-raw-append state
			(format "%s%s\n"
				(pgqa-lisp-indent-string (1+ indent))
				(mapconcat 'format (oref node args) ".")))
  (pgqa-dump-raw-node node-end node state indent))

(defmethod pgqa-dump-raw ((node pgqa-data-type) state indent)
  (pgqa-dump-raw-node node-start node state indent)

  (pgqa-list-slot-to-lisp node state args (1+ indent)))

(defmethod pgqa-dump-raw ((node pgqa-case) state indent)
  (pgqa-dump-raw-node node-start node state indent)

  (if (oref node arg)
      (pgqa-node-slot-to-lisp node state arg (1+ indent)))
  (pgqa-list-slot-to-lisp node state branches (1+ indent))
  (if (slot-boundp node 'else)
      (pgqa-node-slot-to-lisp node state else (1+ indent)))

  (pgqa-dump-raw-node node-end node state indent))

(defmethod pgqa-dump-raw ((node pgqa-case-branch) state indent)
  (pgqa-dump-raw-node node-start node state indent)
  (pgqa-list-slot-to-lisp node state args (1+ indent))
  (pgqa-dump-raw-node node-end node state indent))

(defmethod pgqa-dump-raw ((node pgqa-operator) state indent)
  (pgqa-dump-raw-node node-start node state indent)

  (pgqa-simple-slot-to-lisp node state op (1+ indent))
  (if (slot-boundp node 'args)
      (pgqa-list-slot-to-lisp node state args (1+ indent)))

  (pgqa-dump-raw-node node-end node state indent))

(defmethod pgqa-dump-raw ((node pgqa-node-list) state indent)
  (pgqa-dump-raw-node node-start node state indent)

  (if (slot-boundp node 'args)
      (pgqa-list-slot-to-lisp node state args (1+ indent)))

  (pgqa-dump-raw-node node-end node state indent))

(defmethod pgqa-dump-raw ((node pgqa-target-entry) state indent)
  (pgqa-dump-raw-node node-start node state indent)

  (pgqa-node-slot-to-lisp node state expr (1+ indent))
  (if (slot-boundp node 'alias)
      (pgqa-node-slot-to-lisp node state alias (1+ indent)))

  (pgqa-dump-raw-node node-end node state indent))

(defmethod pgqa-dump-raw ((node pgqa-alias) state indent)
  (pgqa-dump-raw-node node-start node state indent)
  (pgqa-simple-slot-to-lisp node state name (1+ indent))
  (pgqa-dump-raw-node node-end node state indent))

(defmethod pgqa-dump-raw ((node pgqa-alias-arg) state indent)
  (pgqa-dump-raw-node node-start node state indent)
  (pgqa-simple-slot-to-lisp node state var (1+ indent))
  (if (slot-boundp node 'datatype)
      (pgqa-node-slot-to-lisp node state datatype (1+ indent)))
  (pgqa-dump-raw-node node-end node state indent))

(provide 'pgqa-dump-raw)
