;; Copyright (C) 2016 Antonin Houska
;;
;; This file is part of PGQA.
;;
;; PGQA is free software: you can redistribute it and/or modify it under the
;; terms of the GNU General Public License as published by the Free Software
;; Foundation, either version 3 of the License, or (at your option) any later
;; version.

;; PGQA is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
;; details.

;; You should have received a copy of the GNU General Public License qalong
;; with PGQA. If not, see <http://www.gnu.org/licenses/>.

(require 'eieio)

(require 'pgqa-node)

(defcustom pgqa-multiline-query t
  "If non-nil, query will be deparsed in structured format that spans multiple
lines."
  :type 'boolean
  :group 'pgqa
)

(defcustom pgqa-multiline-join t
  "If non-nil, the JOIN operator (including the INNER, LEFT, etc. predicate)
will always start on a new line.

Can only be set if `pgqa-multiline-query' variable is set."
  :type 'boolean
  :group 'pgqa
)

(defcustom pgqa-join-newline t
  "If non-nil, the tables joined (i.e. not only the JOIN operator) start on a
new line.

If either side of the join requires parentheses (i.e. it's a sub-query or a
join), it starts on a new line regardless this setting.

Can only be set if `pgqa-multiline-join' variable is set."
  :type 'boolean
  :group 'pgqa
)

(defcustom pgqa-clause-newline nil
  "If non-nil, each query clause starts on a new line

Can only be set if `pgqa-multiline-query' variable is set."
  :type 'boolean
  :group 'pgqa
)

(defcustom pgqa-clause-item-newline nil
  "If non-nil, each item of query clause starts on a new line

If the item requires parentheses (i.e. it's a sub-query or a join), this
setting is always considered t.

Can only be set if `pgqa-clause-newline' variable is set."
  :type 'boolean
  :group 'pgqa
)

(defcustom pgqa-multiline-operator nil
  "If non-nil, operators will be deparsed in a structured way.

Can only be set if both `pgqa-multiline-query' and `pgqa-multiline-join'
variables are set."
  :type 'boolean
  :group 'pgqa
)

(defcustom pgqa-multiline-case nil
  "If non-nil, each branch of the CASE expression will start on a new line.

Can only be set if both `pgqa-multiline-query' and `pgqa-multiline-join'
variables are set."
  :type 'boolean
  :group 'pgqa
)

(defcustom pgqa-multiline-case-branch nil
  "If non-nil, THEN keyword of the CASE expression will start on a new line.

Can only be set if `pgqa-multiline-case'is set."
  :type 'boolean
  :group 'pgqa
)

(defcustom pgqa-clause-keyword-right nil
  "If non-nil, the clause keyword (SELECT, FROM, WHERE, etc.) will be right
aligned

Can only be set if `pgqa-multiline-query' is set."
  :type 'boolean
  :group 'pgqa)

(defcustom pgqa-print-as t
  "If non-nil, expression alias is denoted by the AS keyword."

  :type 'boolean
  :group 'pgqa)

;; TODO Rename so that it's formatting-specific.
(defun pgqa-check-customizations ()
  "Throw `user-error' if value of any custom variable is illegal."

  (if pgqa-multiline-join
      (if (null pgqa-multiline-query)
	  (user-error "`pgqa-multiline-join' requires
`pgqa-multiline-query'")))

  (if pgqa-join-newline
      (if (null pgqa-multiline-join)
	  (user-error "`pgqa-join-newline' requires `pgqa-multiline-join'")))

  (if pgqa-multiline-case-branch
      (if (null pgqa-multiline-case)
	  (user-error "`pgqa-multiline-case-branch' requires `pgqa-multiline-case'")))

  (if pgqa-multiline-operator
      ;; If the JOIN keyword or the THEN keyword of the CASE expression
      ;; followed an operator expression in the ON clause which is printed out
      ;; in the multiline form, it would become incorrectly indented.
      ;;
      ;; XXX pgqa-multiline-query is redundant here now, but I'm not sure if
      ;; this dependency should be removed. If pgqa-multiline-join became
      ;; unnecessary for pgqa-multiline-operator someday, pgqa-multiline-query
      ;; would have to be added again.
      (if (or (null pgqa-multiline-query) (null pgqa-multiline-join)
	      (null pgqa-multiline-case-branch))
	  (user-error "`pgqa-multiline-operator' requires all `pgqa-multiline-query' and `pgqa-multiline-join' and `pgqa-multiline-case-branch'")))

  (if pgqa-multiline-case
      ;; If the JOIN keyword followed a CASE expression in the ON clausewhich
      ;; is printed out in the multiline form, it would become incorrectly
      ;; indented.
      ;;
      ;; XXX Likewise, pgqa-multiline-query is redundant here.
      (if (or (null pgqa-multiline-query) (null pgqa-multiline-join))
	  (user-error "`pgqa-multiline-case' requires both `pgqa-multiline-query' and `pgqa-multiline-join'")))

  (if pgqa-clause-newline
      (if (null pgqa-multiline-query)
	  (user-error "`pgqa-clause-newline' requires
`pgqa-multiline-query'")))

  (if pgqa-clause-item-newline
      (if (null pgqa-clause-newline)
	  (user-error "`pgqa-clause-item-newline' requires
`pgqa-clause-newline'")))

  (if pgqa-clause-keyword-right
      (if (null pgqa-multiline-query)
	  (user-error "`pgqa-clause-keyword-right' requires
`pgqa-multiline-query'")))
)

;; If buffer-pos is maintained, use the following functions to adjust the
;; start and end position of the node.
(defun pgqa-dump-start (node state)
  (if (oref state buffer-pos)
      ;; Temporarily set the slot to plain number.
      (oset node region (oref state buffer-pos))))

;; The :gui slot can store either pgqa-gui-node or a list of these. This is a
;; convenience routine so that caller does not have to care.
(defun pgqa-dump-start-gui (node state)
  (if (listp node)
      (dolist (i node)
	(pgqa-dump-start i state))
    (pgqa-dump-start node state))
  )

(defun pgqa-dump-end (node state)
  (if (oref state buffer-pos)
      ;; Retrieve the start position stored by pgqa-dump-start and replace it
      ;; with 2-element vector.
      (let* ((start (oref node region))
	     (end (oref state buffer-pos))
	     (region (vector start end)))
	(oset node region region)))
  )

;; Counterpart of pgqa-dump-start-gui.
(defun pgqa-dump-end-gui (node state)
  (if (listp node)
      (dolist (i node)
	(pgqa-dump-end i state))
    (pgqa-dump-end node state))
  )

;; `state' is an instance of `pgqa-deparse-state' class.
;;
;; `indent' determines indentation of the node, relative to (oref node
;; indent).
;;
;; `context' can affect the way node is dumped depending on the surrounding
;; nodes.
(defmethod pgqa-dump ((node pgqa-node) state indent &optional context)
  "Turn a node and its children into string."
  nil)

;; Wrapper function that either calls pgqa-dump or prints out SQL insertion if
;; the node is contained in it.
(defun pgqa-dump-maybe-insertion (node state indent context)
  (let ((insertion)
	(insertion-prev)
	(insertion-key)
	(omit))

    (when context
      (setq insertion (oref node insertion))
      (cl-assert (eq (eieio-object-class-name context)
		     'pgqa-string-to-query-context))
      (setq insertion-prev (oref context last-insertion))

      ;; Prepare the context for the next node.
      (oset context last-insertion insertion)

      (when insertion
	;; If multiple nodes belong to the same insertion, only print the
	;; insertion string once.
	(when (or (null insertion-prev)
		  (null
		   (string=
		    (oref insertion-prev key)
		    (oref insertion key))))
	  (setq insertion-key (oref insertion key))

	  ;; Insert the insertion. Do it in multiple steps so that line can be
	  ;; broken wherever fill-column is reached.
	  (pgqa-deparse-string state "'" indent)
	  (pgqa-deparse-string state "||" indent)
	  (pgqa-deparse-string state insertion-key indent)
	  (pgqa-deparse-string state "||" indent)
	  (pgqa-deparse-string state "'" indent))

	;; Do not dump the contained node itself.
	(setq omit t))
      )

    (if (null omit)
	(if (eq (eieio-object-class node) 'pgqa-query)
	    (pgqa-dump-subquery node state indent context)
	  (pgqa-dump node state indent context)))
    )
  )

;; Metadata to control deparsing of an SQL query and the contained
;; expressions.
;;
;; Note: the simplest way to control (not) breaking line in front of a
;; subquery is to create a separate instance of this class for the subquery.
;;
;; XXX If adding a new field, check if it's subject to backup / restore by
;; pgqa-dump method of pgqa-operator class. (Is this backup / restore worth
;; separate functions?)
;;
;; TODO Rename to pgqa-dump-state ?
(defclass pgqa-deparse-state ()
  (
   (indent
    :initarg :indent
    ;; The indentation passed to `pgqa-dump', `pgqa-deparse-string' and
    ;; subroutines is relative to this.
    :documentation "Indentation level of the top level query.")

   (indent-top-expr
    :initarg :indent-top-expr
    :documentation "Indentation, relative to `indent' above, of top level
expression, e.g. the text following top-level keyword. See
`pgqa-deparse-top-keyword' for details.")

   (next-column
    :initarg :next-column
    :documentation "Column at which the following node (or its leading space)
should start.")

   (next-space
    :initarg :next-space
    :documentation "The width of the next space to be printed out.")

   (line-empty
    :initarg :line-empty
    :documentation "Is the current line empty or contains only whitespace?")

   (buffer-pos
    :initarg :buffer-pos
    :documentation "Position in the output buffer at which the next string will start."
    )

   (result
    :initarg :result
    :documentation "String to which each node appends its textual
representation.")
   )

  "State of the deparsing process."
  )

;; init-col-src is column (in addition to the indentation) at which the source
;; query starts.
(defun pgqa-init-deparse-state (indent init-col-src line-empty buffer-pos)
  "Initialize instance of `pgqa-deparse-state'."

  (let ((indent-width (* indent tab-width)))
    (make-instance 'pgqa-deparse-state
		   :indent indent
		   :indent-top-expr 0
		   :next-column (+ indent-width init-col-src)
		   :next-space 0
		   :line-empty t
		   :buffer-pos buffer-pos
		   :result (make-string indent-width 32))))

(defmethod pgqa-deparse-state-get-attrs ((state pgqa-deparse-state))
  "Return the slots of `pgqa-deparse-state' as an association list."
  (let ((result)
	(item))
    (dolist (key (object-slots state) result)
      (setq item (list key (slot-value state key)))
      (push item result))))

(defmethod pgqa-deparse-state-set-attrs ((state pgqa-deparse-state) slots)
  "Set slots of `pgqa-deparse-state' to values extracted by
`pgqa-deparse-state-get-attrs'."
  (dolist (slot slots)
    (set-slot-value state (car slot) (car (cdr slot)))))

;; TODO Check if call is always followed by (oset state next-space 0). If so,
;; incorporate it into the method.
(defmethod pgqa-deparse-newline ((state pgqa-deparse-state) indent)
  "Adjust deparse state so that deparsing continues at a new line, properly
indented."

  (let* ((indent-loc (+ indent (oref state indent)))
	 (indent-width (* indent-loc tab-width))
	 (result (oref state result)))
    (setq result (concat result "\n"))
    (setq result (concat result
			 (make-string indent-width 32)))
    (oset state result result)
    (oset state next-column indent-width)

    ;; buffer-pos might need to account for the strings added.
    (if (oref state buffer-pos)
	(oset state buffer-pos (+ (oref state buffer-pos) 1 indent-width)))

    (oset state line-empty t))
  )

(defmethod pgqa-deparse-space ((state pgqa-deparse-state))
  "Write space to deparsing output."
  (let ((space (oref state next-space)))
    (oset state result
	  (concat (oref state result)
		  (make-string space 32)))
    (oset state next-column (+ (oref state next-column) space))

    ;; buffer-pos might need to account for the strings added.
    (if (oref state buffer-pos)
	(oset state buffer-pos (+ (oref state buffer-pos) space))))

  ;; Restore the default value of next-space.
  (oset state next-space 1))

;; Prepare for insertion of `str', i.e. add line break or space, as
;; needed.
(defmethod pgqa-deparse-string-prepare ((state pgqa-deparse-state) str indent)
  (let ((col-incr 0)
	(space (oref state next-space)))
    (if (> space 0)
	(setq col-incr (1+ col-incr)))
    (setq col-incr (+ col-incr (string-width str)))

    ;; Zero space currently can't be broken.
    ;;
    ;; TODO Consider if there are special cases not to subject to this
    ;; restriction, and maybe introduce custom variable that allows breaking
    ;; even the zero space.
    (when (and fill-column (> space 0)
	       (> (+ (oref state next-column) col-incr) fill-column))
      (pgqa-deparse-newline state indent)
      ;; No space (in addition to indentation) after newline.
      (oset state next-space 0))

    (pgqa-deparse-space state))
  )

(defmethod pgqa-deparse-string ((state pgqa-deparse-state) str indent)
  "Write arbitrary string to deparsing output."

  (pgqa-deparse-string-prepare state str indent)

  ;; In some cases we stick the next string to the current one w/o space
  ;; (which currently makes newline impossible - see
  ;; pgqa-deparse-string-prepare).
  (if (or
       (string= str "(") (string= str "["))
      (oset state next-space 0))

  (oset state result (concat (oref state result) str))

  (let ((str-width (string-width str)))
    (oset state next-column (+ (oref state next-column) str-width))

    ;; buffer-pos might need to account for the strings added.
    (if (oref state buffer-pos)
	(oset state buffer-pos (+ (oref state buffer-pos) str-width))))

  ;; clear line-empty if the string contains non-whitespace character.
  (if (string-match "\\S-" str)
      (oset state line-empty nil)))

;; Top-level keyword might deserve special attention, e.g. adding tabs between
;; itself and the following expression.
;;
;; `first' indicates that this is the first top-level keyword of the whole
;; query.
(defmethod pgqa-deparse-top-keyword ((state pgqa-deparse-state) keyword first)
  "Dump top-level keyword (SELECT, INSERT, FROM, WHERE, etc.)"

  (let ((first-offset 0)
	(nspaces))

    ;; For the structured output, all top-level keywords except for the first
    ;; one need line break.
    (when (and pgqa-multiline-query (null first))
      (pgqa-deparse-newline state 0)
      ;; No space in front of the keyword, even if the keyword does not cause
      ;; line break itself.
      (oset state next-space 0))

    ;; The first line of the deparsed query may be indented more than the rest
    ;; (see indent-estimate in pgqa-deparse).
    (if (and pgqa-multiline-query first)
	(setq first-offset
	      (- (oref state next-column)
		 (* (oref state indent) tab-width))))

    (if (and pgqa-multiline-query
	     ;; Make sure the clause is correctly indented.
	     ;;
	     ;; Right alignment of the clause keyword is also a reason to
	     ;; compute the space, even though the clause might end up on a
	     ;; new line. In such a case we'll simply put the space left from
	     ;; the keyword.
	     (or
	      (null pgqa-clause-newline)
	      pgqa-clause-keyword-right))

	(let* ((indent-top-expr (oref state indent-top-expr)))
	  (setq nspaces (- (* indent-top-expr tab-width) (string-width keyword)))

	  ;; Anything wrong about computation of indent-top-expr?
	  (if (and (null pgqa-clause-newline) (< nspaces 1))
	      (error "indent-top-expr is too low"))

	  ;; No extra space if the next text would exceed fill-column.
	  (if (and
	       fill-column
	       (>= (+ (oref state next-column) nspaces) fill-column))
	      (setq nspaces 1))

	  ;; Shorten the space so that the expression (most likely column
	  ;; list) starts at the same position as the other expressions
	  ;; (e.g. table list).
	  (when (> first-offset 0)
	    (setq nspaces (- nspaces first-offset))
	    ;; Make sure at least one space is left.
	    (if (< nspaces 1)
		(setq nspaces 1)))))

    ;; If the clause keyword should be right-aligned, put the space in front
    ;; of it. but leave 1 to be written between the keyword and the actual
    ;; clause expression.
    (when (and nspaces (> nspaces 1) pgqa-clause-keyword-right)
      ;; If the clause should appear on the same line as the keyword, retain
      ;; one space to separate them. If the clause starts at a new line, use
      ;; the whole nspaces for the right alignment.
      (let ((i (if pgqa-clause-newline 0 1)))
	(while (< i nspaces)
	  (pgqa-deparse-space state)
	  (setq nspaces (1- nspaces)))))

    (pgqa-deparse-string state keyword
			 (if pgqa-multiline-query 1 0))

    ;; Separate the clause from the keyword.
    (if nspaces
	(oset state next-space nspaces))

  (if pgqa-clause-newline
      ;; Avoid the single space that pgqa-dump of pgqa-operator class puts in
      ;; front of operators.
      (oset state next-space 0)))
  )

;; In this subclass the function does not need the `indent' argument - the
;; base indentation is controlled by (oref state indent). (EIEIO does not
;; allow omitting the argument altogether.)
(defmethod pgqa-dump ((node pgqa-query) state indent context)
  "Turn query into a string."

  (pgqa-dump-start node state)
  ;; For mutiline output, compute the first column for expressions.
  (if pgqa-multiline-query
      (if pgqa-clause-newline
	  ;; No need to separate the keyword from the clause using a space, so
	  ;; no need to adjust indent-top-expr.
	  (oset state indent-top-expr 1)
	;; The additional space can affect indent-top-expr.
	(progn
	  ;; TODO Put the following into a separate function
	  (let ((top-clauses)
		(max-width 0)
		(indent-expr)
		(kind (oref node kind)))
	    (if (oref node target-expr)
		(push "SELECT" top-clauses))

	    (if (oref node target-table)
		(if (string= kind "UPDATE")
		    (push "UPDATE" top-clauses)
		  (if (string= kind "INSERT")
		      (push "INSERT INTO" top-clauses))))

	    (if (oref node from-expr)
		(let ((fe (oref node from-expr)))
		  (if (> (length (oref fe from-list)) 0)
		      (push "FROM" top-clauses))
		  (if (slot-boundp fe 'qual)
		      (push "WHERE" top-clauses))
		  ))

	    (if (oref node group-by)
	      (push "GROUP BY" top-clauses))
	    (if (oref node having)
	      (push "HAVING" top-clauses))
	    (if (oref node order-by)
	      (push "SORT BY" top-clauses))

	    ;; Find out the maximum length.
	    (dolist (i top-clauses)
	      (let ((width (string-width i)))
		(if (> width max-width)
		    (setq max-width width))))

	    ;; At least one space must follow.
	    (setq max-width (1+ max-width))

	    ;; Round up to the nearest multiple of tab-width.
	    (setq max-width
		  (+ max-width
		     (- tab-width (% max-width tab-width))))

	    (oset state indent-top-expr (/ max-width tab-width))))
	)
    )

  (let ((indent-clause))
    (if pgqa-clause-newline
	(setq indent-clause 1)
      ;; Extra tab might have been added in front of the clause (to ensure
      ;; that all clauses of the query start at the same position), so all
      ;; lines of the clause must start at that position.
      (setq indent-clause (oref state indent-top-expr)))

    (if (string= (oref node kind) "SELECT")
	(let ((te (oref node target-expr)))
	  (pgqa-deparse-top-keyword state "SELECT" t)

	  ;; Enforce line break if necessary.
	  ;;
	  ;; TODO The same for ORDER BY, WINDOW, LIMIT, etc.
	  (if pgqa-clause-newline
	      (pgqa-deparse-newline state indent-clause))

	  (pgqa-dump te state indent-clause context)))

    (if (string= (oref node kind) "UPDATE")
	(let ((tt (oref node target-table))
	      (te (oref node target-expr)))
	  (pgqa-deparse-top-keyword state "UPDATE" t)
	  (if pgqa-clause-newline
	      (pgqa-deparse-newline state indent-clause))
	  (pgqa-dump tt state indent-clause context)
	  (pgqa-deparse-top-keyword state "SET" nil)
	  (if pgqa-clause-newline
	      (pgqa-deparse-newline state indent-clause))
	  (pgqa-dump te state indent-clause context)))

    (if (string= (oref node kind) "INSERT")
	(let ((tt (oref node target-table))
	      (te (oref node target-expr))
	      (ic (oref node insert-cols)))
	  (pgqa-deparse-top-keyword state "INSERT INTO" t)
	  (if pgqa-clause-newline
	      (pgqa-deparse-newline state indent-clause))
	  (pgqa-dump tt state indent-clause context)
	  (when ic
	    (oset state next-space 0)
	    (pgqa-deparse-string state "(" indent)
	    (pgqa-dump ic state indent-clause context)
	    (oset state next-space 0)
	    (pgqa-deparse-string state ")" indent))
	  (if te
	      (pgqa-deparse-top-keyword state "SELECT" nil))
	  (if pgqa-clause-newline
	      (pgqa-deparse-newline state indent-clause))
	  (pgqa-dump te state indent-clause context)
	  ))

    (if (oref node from-expr)
	(let ((from-expr (oref node from-expr)))
	  ;; Update may or may not have FROM clause.
	  (pgqa-dump from-expr state 0 context)))

    (if (oref node group-by)
      (pgqa-dump (oref node group-by) state 0 context))
    (if (oref node having)
      (pgqa-dump (oref node having) state 0 context))
    (if (oref node order-by)
      (pgqa-dump (oref node order-by) state 0 context))
    )
  (pgqa-dump-end node state))

;; Find out if from-list-entry needs to be parenthesized.
;;
;; join-rhs is t if the node is join RHS, nil if its join LHS or
;; an item of the FROM-list.
(defun pgqa-from-list-entry-needs-parens (node join-rhs)
  (if (eq (eieio-object-class node) 'pgqa-from-list-entry)
      (let ((is-join (= (length (oref node args)) 2))
	    (is-query (string= (oref node kind) "query")))
	;; Query needs the parentheses always, join only if it's on the right side
	;; of another join or if it has an alias.
	(or is-query (and is-join (or join-rhs (slot-boundp node 'alias))))))
  )

;; Return the FROM-list subquery or nil if the entry is something else (table,
;; join, etc.).
(defun pgqa-get-from-list-suquery (fe)
  (if (and (eq (eieio-object-class fe) 'pgqa-from-list-entry)
	    (string= (oref fe kind) "query"))
      (car (oref fe args)))
  )

(defmethod pgqa-dump ((node pgqa-from-expr) state indent context)
  (pgqa-dump-start node state)

  (let ((from-list (oref node from-list))
	(indent-clause))

    ;; See the related comment in pgqa-dump method of pgqa-query class.
    (if pgqa-clause-newline
	(setq indent-clause (1+ indent))
      (setq indent-clause (oref state indent-top-expr)))

    ;; INSERT, UPDATE or DELETE statement can have the list empty.
    (if (> (length from-list) 0)
	(progn
	  (pgqa-deparse-top-keyword state "FROM" nil)

	  ;; (if pgqa-clause-newline
	  ;;     (pgqa-deparse-newline state indent-clause))

	  (let* ((i 0)
		 (fe-query)
		 (newline)
		 (parens)
		 (is-join))
	    (dolist (item from-list)
	      (if (> i 0)
		  (progn
		    (oset state next-space 0)
		    (pgqa-deparse-string state "," indent-clause)
		    ;; In general, comma should be followed by a space. As for
		    ;; special cases, we'll set it to zero explicitly below.
		    (if (null pgqa-join-newline)
			(oset state next-space 1))))

	      (setq parens (pgqa-from-list-entry-needs-parens item nil))
	      (setq is-join (= (length (oref item args)) 2))
	      (setq fe-query (pgqa-get-from-list-suquery item))

	      ;; XXX Can we do anything better than breaking the line if
	      ;; parens is t but pgqa-clause-item-newline is nil?
	      ;;
	      ;; TODO Consider newline even if pgqa-join-newline is nil but
	      ;; the next entry is a join that starts with left parenthesis
	      ;; (i.e. the join needs the parentheses itself, or its left-most
	      ;; argument does).
	      (setq newline
		    (or
		     (and (= i 0) pgqa-clause-newline)
		     (and
		      (> i 0)
		      (or (and parens pgqa-multiline-query) pgqa-clause-item-newline
			  (and is-join pgqa-join-newline))
		      (null (oref state line-empty)))))

	      (if newline
		  (pgqa-deparse-newline state indent-clause))

	      ;; Set next-space to 0 to avoid unnecessary space, or even
	      ;; newline in the case of a join.
	      (if (> (oref state next-space) 0)
		  (if (or parens is-join pgqa-clause-item-newline)
		      (progn
			;; Only apply the space if non-empty line continues
			;; here.
			(if (null (oref state line-empty))
			    (pgqa-deparse-space state))
			(oset state next-space 0))
		    )
		)

	      (if parens
		  (pgqa-deparse-string state "(" indent))

	      (if fe-query
		  (pgqa-dump-subquery (car (oref item args))
					     state indent-clause context)
		(pgqa-dump-maybe-insertion item state indent-clause context))

              (if parens
                  (progn
                    (oset state next-space 0)
                    (pgqa-deparse-string state ")" indent-clause)))

	      (if (slot-boundp item 'alias)
		  (pgqa-dump (oref item alias) state indent-clause context))
	      (setq i (1+ i)))
	    )
	  )
      )
    )

  (if (slot-boundp node 'qual)
      (let ((indent-clause))
	;; Like above. XXX Should the whole body of the function be wrapped in
	;; an extra "let" construct, which initializes the variable only
	;; once?)
	(if pgqa-clause-newline
	    (setq indent-clause (1+ indent))
	  (setq indent-clause (oref state indent-top-expr)))

	;; `space' should be up-to-date as the FROM clause is mandatory.
	(pgqa-deparse-top-keyword state "WHERE" nil)
	(if pgqa-clause-newline
	    (pgqa-deparse-newline state indent-clause))
	(pgqa-dump (oref node qual) state indent-clause context)))
  (pgqa-dump-end node state))

(defmethod pgqa-dump ((node pgqa-from-list-entry) state indent context)
  "Print out FROM list entry (table, join, subquery, etc.)."

  (pgqa-dump-start node state)
  (let* ((args (oref node args))
	 (nargs (length args))
	 (is-join (= nargs 2))
	 (arg (car args))
	 (fe-query)
	 (parens)
	 (newline)
	 ;; If this happens to be a join nested in another one, that upper
	 ;; join might use next-space to indicate that it already took care of
	 ;; the newline.
	 (no-space (and is-join (= (oref state next-space) 0))))

    (cl-assert (or (= nargs 1) (= nargs 2)))

    (setq parens (pgqa-from-list-entry-needs-parens arg nil))
    (if (null is-join)
	(progn
	  ;; If the node is query, it'll be handled by upper node, which is
	  ;; responsible for parentheses and line breaks.
	  (setq fe-query (pgqa-get-from-list-suquery node))
	  (if (null fe-query)
	      (pgqa-dump arg state indent context)))
      (progn
	;; arg is LHS of the join.
	;;
	;; XXX Can we do anything smarter than breaking the line if parens is
	;; t but pgqa-join-newline is nil?
	(setq newline
	      (and
	       pgqa-multiline-query
	       (or parens pgqa-join-newline)
	       ;; Only break the line if it hasn't just happened for any
	       ;; reason.
	       (null (oref state line-empty))
	       ;; The purpose of no-space is explained above.
	       (null no-space)))

	(if newline
	    (progn
	      ;; TODO Shouldn't pgqa-deparse-newline set next-space to zero?
	      (pgqa-deparse-newline state indent)
	      (oset state next-space 0)))

	(if parens
	    (progn
	      ;; There should be no space in front of the next character, but
	      ;; we also set next-space to zero to indicate that nested entry
	      ;; should not be preceded by a newline.
	      (oset state next-space 0)

	      (pgqa-deparse-string state "(" indent)))

	(setq fe-query (pgqa-get-from-list-suquery arg))

	(if fe-query
	    (pgqa-dump-subquery fe-query state indent context)
	  (pgqa-dump arg state indent context))

	(if parens
	    (progn
	      (oset state next-space 0)
	      (pgqa-deparse-string state ")" indent)))

	;; Print the alias separate so that it does not appear inside the
	;; parentheses.
	(if (slot-boundp arg 'alias)
	    (pgqa-dump (oref arg alias) state indent context))
	)
      )

    (if is-join
	(progn
	  (if pgqa-multiline-join
	      (progn
		(oset state next-space 0)
		(pgqa-deparse-newline state indent)))

	  (let ((kind (oref node kind)))
	    (if kind
		(pgqa-deparse-string state (upcase kind) indent)))
	  (pgqa-deparse-string state "JOIN" indent)

	  ;; The right side of a join might require parentheses.
	  (let* ((arg (nth 1 args)))
	    (setq parens (pgqa-from-list-entry-needs-parens arg t))
	    ;; XXX Can we do anything smarter than breaking the line if parens
	    ;; is t but pgqa-join-newline is nil?
	    (setq newline
		  (and pgqa-multiline-query
		       (or parens pgqa-join-newline)
		       ;; Only break the line if it hasn't just happened for
		       ;; any reason.
		       (null (oref state line-empty))))

	    (if newline
		(progn
		  (oset state next-space 0)
		  (pgqa-deparse-newline state indent)))

	    (if parens
		(pgqa-deparse-string state "(" indent))

	    (setq fe-query (pgqa-get-from-list-suquery arg))

	    (if fe-query
		(pgqa-dump-subquery fe-query state indent context)
	      (pgqa-dump arg state indent context))

	    (if parens
		(progn
		  (oset state next-space 0)
		  (pgqa-deparse-string state ")" indent)))

	    ;; pgqa-dump-subquery couldn't print the alias because thus
	    ;; it'd appear inside the parentheses.
	    (if (slot-boundp arg 'alias)
		(pgqa-dump (oref arg alias) state indent context)))

	  (pgqa-deparse-string state "ON" indent)

	  (pgqa-dump (oref node qual) state
		     (if pgqa-multiline-join (1+ indent) indent) context)))
    )
  (pgqa-dump-end node state))

;; Dump query in the FROM list or in an expression.
;;
;; TODO Consider custom variable that controls whether parentheses are on the
;; same lines the query starts and ends respectively.
(defun pgqa-dump-subquery (query state indent context)
  (let ((state-loc state))
    (if pgqa-multiline-query
	;; Use a separate state to print out query.
	;;
	;; init-col-src of 1 stands for the opening parenthesis.
	(progn
	  (setq state-loc (pgqa-init-deparse-state
			   (+ (oref state indent) indent) 1 t
			   (oref state buffer-pos)))
	  (oset state-loc next-column (oref state next-column))
	  (oset state-loc result (oref state result))))

    (pgqa-dump query state-loc 0 context)
    (oset state result (oref state-loc result))

    (if pgqa-multiline-query
	(progn
	  (oset state next-column (oref state-loc next-column))
	  (oset state buffer-pos (oref state-loc buffer-pos)))))

  (oset state next-space 0)
)

(defmethod pgqa-dump ((node pgqa-from-list-entry-alias) state indent context)
  "Turn alias into a string."
  (if pgqa-print-as
      (pgqa-deparse-string state "AS" indent))
  (pgqa-dump-start node state)
  (pgqa-dump (oref node name) state indent context)
  ;; Print out argument list if there is some.
  (when (slot-boundp node 'cols)
    ;; No space in front of the left parenthesis.
    (oset state next-space 0)
    (pgqa-deparse-string state "(" indent)
    (pgqa-dump (oref node cols) state indent context)
    ;; No space in front of the right parenthesis.
    (oset state next-space 0)
    (pgqa-deparse-string state ")" indent))
  (pgqa-dump-end node state))

(defmethod pgqa-dump ((node pgqa-top-clause) state indent context)
  (pgqa-dump-start node state)

  (let* ((indent-clause)
	 (cn (eieio-object-class-name node))
	 (kwd)
	 (expr (oref node expr)))

    (setq kwd
	  (if (eq cn 'pgqa-group-clause)
	      "GROUP BY"
	    (if (eq cn 'pgqa-having-clause)
		"HAVING"
	      (if (eq cn 'pgqa-sort-clause)
		  "ORDER BY"
		(error
		 (format "Unrecognized top level expression class %s." cn))))))

    (pgqa-deparse-top-keyword state kwd nil)

    ;; See the related comment in pgqa-dump method of pgqa-query class.
    (if pgqa-clause-newline
	(setq indent-clause (1+ indent))
      (setq indent-clause (oref state indent-top-expr)))

    (if pgqa-clause-newline
	(pgqa-deparse-newline state indent-clause))

    (pgqa-dump expr state indent-clause context))
  (pgqa-dump-end node state))

(defmethod pgqa-dump ((node pgqa-func-call) state indent context)
  "Print out function call"

  ;; Dump the function name and use pgqa-dump-start / pgqa-dump-end to mark
  ;; set the region of the operator string. This is needed so we can
  ;; eventually assign face to the string.
  (pgqa-dump-start (nth 0 (oref node gui)) state)
  (pgqa-dump (oref node name) state indent context)
  (pgqa-dump-end (nth 0 (oref node gui)) state)

  ;; No space between function name and the parenthesis.
  (oset state next-space 0)

  ;; Adjust the GUI node of the left parenthesis.
  (pgqa-dump-start (nth 1 (oref node gui)) state)
  (pgqa-deparse-string state "(" indent)
  (pgqa-dump-end (nth 1 (oref node gui)) state)

  (let ((args (oref node args)))
    (if (> (length args) 0)
	(pgqa-dump args state indent context)))
  ;; Likewise, no space after.
  (oset state next-space 0)

  ;; Adjust the GUI node of the right parenthesis.
  (pgqa-dump-start (nth 2 (oref node gui)) state)
  (pgqa-deparse-string state ")" indent)
  (pgqa-dump-end (nth 2 (oref node gui)) state)
)


(defmethod pgqa-dump ((node pgqa-string) state indent context)
  "Dump a string constant."

  (pgqa-dump-start node state)

  (let ((str (oref node value)))
    ;; If we're converting query to SQL string, string constants must be
    ;; quoted.
    (if (and context
	     (eq (eieio-object-class-name context)
		 'pgqa-string-to-query-context))
	(setq str (format "'%s'" str)))
    (pgqa-deparse-string state str indent))

  (pgqa-dump-end node state))

(defmethod pgqa-dump ((node pgqa-number) state indent context)
  "Turn number into a string."

  (pgqa-dump-start node state)

  (let ((str (oref node value)))
    (pgqa-deparse-string state str indent))

  (pgqa-dump-end node state))

(defmethod pgqa-dump ((node pgqa-obj) state indent context)
  "Turn an SQL object into a string."

  (pgqa-dump-start node state)
  (let ((str (mapconcat 'format (oref node args) ".")))
    (pgqa-deparse-string state str indent))
  (pgqa-dump-end node state))

(defmethod pgqa-dump ((node pgqa-data-type) state indent context)
  "Turn data type into a string."

  (pgqa-dump-start node state)
  (dolist (i (oref node args))
    (pgqa-dump i state indent context))
  (pgqa-dump-end node state))

(defmethod pgqa-dump ((node pgqa-case) state indent context)
  "Print out the CASE expression."

  (pgqa-dump-start node state)

  ;; The multi-line output should start exactly at the indentation which is
  ;; determined by the indent argument, so new line is needed.
  (when (and pgqa-multiline-case (null (oref state line-empty)))
    (pgqa-deparse-newline state indent)
    (oset state next-space 0))

  (pgqa-deparse-string state "CASE" indent)

  (let ((indent-2 indent))
    (if pgqa-multiline-case
      (setq indent-2 (1+ indent-2)))

    (if (oref node arg)
	;; Even the argument deserves extra indentation, in case it spans
	;; multiple lines.
	(pgqa-dump (oref node arg) state indent-2 context))

    (dolist (branch (oref node branches))
      (when pgqa-multiline-case
	(pgqa-deparse-newline state indent-2)
	(oset state next-space 0))
      (pgqa-dump branch state indent-2 context))

    (when (slot-boundp node 'else)
      (when pgqa-multiline-case
	(pgqa-deparse-newline state indent-2)
	(oset state next-space 0))
      (pgqa-deparse-string state "ELSE" indent-2)
      (pgqa-dump (oref node else) state indent-2 context))
    )
  (when pgqa-multiline-case
    (pgqa-deparse-newline state indent)
    (oset state next-space 0))
  (pgqa-deparse-string state "END" indent)
  (pgqa-dump-end node state))

(defmethod pgqa-dump ((node pgqa-case-branch) state indent context)
  "Print out a single branch (WHEN ... THEN ...) of the CASE expression."

  (pgqa-dump-start node state)
  (let* ((args (oref node args))
	 (when-expr (car args))
	 (then-expr (car (cdr args)))
	 (indent-2 indent))
    (pgqa-deparse-string state "WHEN" indent)

    ;; The arguments should get extra indentation, if they span multiple
    ;; lines. (Not sure it looks nice if pgqa-multiline-case is sufficient to
    ;; cause this extra indentation.)
    (if pgqa-multiline-case-branch
	(setq indent-2 (1+ indent-2)))

    (pgqa-dump when-expr state indent-2 context)

    (when pgqa-multiline-case-branch
      (pgqa-deparse-newline state indent)
      (oset state next-space 0))

    (pgqa-deparse-string state "THEN" indent)
    (pgqa-dump then-expr state indent-2 context))

  ;; If the WHEN and THEN clauses are separated by a newline, the whole CASE
  ;; expression looks better if an extra line break is added in front of the
  ;; next branch (or in front of the ELSE clause).
  (if pgqa-multiline-case-branch
      (pgqa-deparse-newline state 0))
  (pgqa-dump-end node state))

(defun pgqa-child-op-needs-parens (node arg)
  "Find out if argument of an operator node should be parenthesized."
  (if
      (or
       ;; Query as an argument should always be parenthesized.
       (eq (eieio-object-class arg) 'pgqa-query)

       ;; Sublink can have either query or an (array) expression as an
       ;; argument, both of which should always be parenthesized.
       (eq (eieio-object-class node) 'pgqa-sublink))
      t
    (let ((prec (oref node prec))
	  (prec-child
	   (if (object-of-class-p arg pgqa-operator)
	       (oref arg prec))))
      (and prec prec-child (> prec prec-child))))
  )

(defun pgqa-is-multiline-operator (node)
  "Should the argument be printed in structured way?"

  (if (and pgqa-multiline-operator
	   (or
	    (object-of-class-p node 'pgqa-operator)
	    (and
	     pgqa-multiline-case
	     (object-of-class-p node 'pgqa-case))))

      ;; List is not a multi-line operator as such, although its arguments
      ;; can be.
      (null (object-of-class-p node 'pgqa-node-list))
    )
  )

;; indent relates to the operator, not argument.
(defun pgqa-indent-operator-first-argument (state indent arg-idx)
  "Prepare position for the first argument of a multiline operator."

  (let* ((s (oref state result))
	 (i (1- (length s))))
    (if
	;; No duplicate newline if we already have one.
	(and pgqa-clause-newline (= arg-idx 0) (oref state line-empty))
	(let ((indent-extra
	       (-
		;; indent argument of the function is relative to (oref state
		;; indent), so compute "absolute value".
		(+ (oref state indent) indent)
		(/ (oref state next-column) tab-width)))
	      (indent-extra-spaces))
	  ;; indent is for the operator, so add 1 more level for the argument.
	  (setq indent-extra (1+ indent-extra))
	  (setq indent-extra-spaces (* indent-extra tab-width))
	  (oset state result
		(concat (oref state result)
			(make-string indent-extra-spaces 32)))
	  (if (oref state buffer-pos)
	      (oset state buffer-pos (+ (oref state buffer-pos)
					indent-extra-spaces))))
      (progn
	(pgqa-deparse-newline state (1+ indent))
	(oset state next-space 0))))
  )

(defmethod pgqa-dump ((node pgqa-operator) state indent context)
  "Turn operator expression into a string."

  (pgqa-dump-start node state)
  (let* ((args (oref node args))
	 (nargs (length args))
	 (op (oref node op))
	 (is-list (object-of-class-p node 'pgqa-node-list))
	 (is-unary
	  (null
	   (or (cdr args) is-list)))
	 (is-cast (string= op "::"))
	 (i 0)
	 (multiline
	  (and
	   pgqa-multiline-operator
	   (pgqa-is-multiline-operator node)))
	 (arg-multiline-prev))

    (dolist (arg args)
      (let* ((parens (pgqa-child-op-needs-parens node arg))
	     (state-backup)
	     (arg-is-operator (object-of-class-p arg 'pgqa-operator))
	     (arg-is-list (object-of-class-p arg 'pgqa-node-list))
	     (arg-is-te (eq (eieio-object-class arg) 'pgqa-target-entry))
	     ;; FROM list entry?
	     (arg-is-fe (eq (eieio-object-class arg) 'pgqa-from-list-entry))
	     (arg-is-query (eq (eieio-object-class arg) 'pgqa-query))
	     (arg-multiline)
	     (indent-arg indent)
	     (gui (oref node gui)))

	(if arg-is-te
	    (setq arg-multiline (pgqa-is-multiline-operator (oref arg expr)))
	  (setq arg-multiline (pgqa-is-multiline-operator arg)))

	(if (or (and is-unary (null (oref node postfix))) (> i 0))
	    (let ((omit-space))

	      ;; Never put space in front of comma or cast operator.
	      (setq omit-space (or is-list is-cast))

	      ;; Should each arg start at a new line?
	      (when multiline
		(pgqa-deparse-newline state indent)
		(setq omit-space t))

	      (if omit-space
	      	  (oset state next-space 0))

	      ;; BETWEEN-AND operator has 2 keywords, and "gui" is a 2-element
	      ;; list containing one node per keyword.
	      (if (equal op "BETWEEN")
		  (if (= i 1)
		      (setq gui (car gui))
		    (when (= i 2)
		      (setq op "AND")
		      (if gui
			  (setq gui (car (cdr gui))))
		      )
		    )
		)

	      ;; Use pgqa-dump-start / pgqa-dump-end to mark set the region of
	      ;; the operator string. This is needed so we can eventually
	      ;; assign face to the string.
	      (if gui
		  (pgqa-dump-start gui state))

	      (pgqa-deparse-string state op indent)

	      ;; Cast operator (::) is special in that it should not be
	      ;; separated from the target type by space.
	      (if is-cast
		  (oset state next-space 0))

	      (if gui
		  (pgqa-dump-end gui state)))

	  )

	;; Ensure correct initial position for the argument output in case the
	;; operator spans multiple lines.
	(if multiline
	    (if (or
		 ;; "(" should appear on a new line, indented as the argument
		 ;; would be if there were no parentheses. (The argument itself
		 ;; will eventually be given extra indentation.)
		 parens

		 (and
		  ;; If the arg is also an operator, it'll take care of the
		  ;; initial line preparation itself.
		  (null arg-is-operator)

		  ;; If the current operator (i.e. not the argument) is unary
		  ;; prefix (not enclosed in parentheses) - it looks better if
		  ;; both operator and argument end up on the same
		  ;; line. Again, if the argument of the unary operator is
		  ;; operator itself (which includes parentheses), it'll take
		  ;; care of the preparation itself.
		  (or (> (length args) 1) (oref node postfix))
		  )
		 )
		(pgqa-indent-operator-first-argument state indent i)
	      )

	  (when
	      (or
	       (and is-list

		    ;; Multi-line argument takes care of the initial line
		    ;; preparation itself.
		    (null arg-multiline)

		    (or
		     ;; If an "ordinary" expression follows a multi-line
		     ;; operator within comma operator (e.g. SELECT list),
		     ;; break the line so that the multi-line operator does not
		     ;; share even a single line with the current argument.
		     ;;
		     ;; TODO Consider a custom variable to switch this behavior
		     ;; on / off.
		     arg-multiline-prev

		     ;; Definitely break the line if user requires each target
		     ;; list / from list entry to begin on a new line.
		     ;;
		     ;; Do nothing for i = 0 because pgqa-clause-item-newline
		     ;; requires pgqa-clause-newline to be set, which takes
		     ;; care of the first entry. Only take action if the comma
		     ;; is a target list of FROM list (i.e. do not affect
		     ;; things like function argument list).
		     (and pgqa-clause-item-newline (> i 0)
			  (or arg-is-te arg-is-fe))))

	       ;; Another reason to break the line is a query in an expression
	       ;; (e.g. IN). If we didn't break the line here, the first line
	       ;; (typically SELECT) whould be much more indented than the
	       ;; following lines.
	       (and pgqa-multiline-query arg-is-query))

	    (pgqa-deparse-newline state indent)
	    (oset state next-space 0))
	  )

	(when parens
	  ;; In case we're not at the start of a new line, so make sure that
	  ;; there's no space between the sublink keyword and the left
	  ;; parenthesis. XXX Consider customization variable that allows such
	  ;; a space.
	  (if (eq (eieio-object-class node) 'pgqa-sublink)
	      (oset state next-space 0))
	  (pgqa-deparse-string state "(" indent)
	  ;; No space, whatever follows "(".
	  (oset state next-space 0))

	;; Comma needs special treatment because it doesn't look nice if it's
	;; the first character on a line.
	;;
	;; Backup the slots but keep the instance so that callers still see
	;; our changes.
	(setq state-backup (pgqa-deparse-state-get-attrs state))

	;; Serialize the argument now, giving it additional indentation if
	;; user wants the output structured.
	(let ((indent-extra 0))
	  (when multiline
	    (setq indent-extra (1+ indent-extra))
	    ;; The extra indentation due to parentheses, mentioned above. It
	    ;; only applies to operators, not to subqueries.
	    (if (and parens (null arg-is-query))
		(setq indent-extra (1+ indent-extra))))
	  (setq indent-arg (+ indent indent-extra))

	  ;; Operator can be contained in an SQL string insertion.
	  (pgqa-dump-maybe-insertion arg state indent-arg context))

	(when
	    ;; If the argument has reached fill-column, line should have
	    ;; broken in front of the argument. So restore the previous state
	    ;; and dump the argument again, with fill-column temporarily
	    ;; decreased by one. That should make the argument appear on the
	    ;; new line too.
	    (and
	     (>= (oref state next-column) fill-column)
	     (< i (1- nargs)))

	  (pgqa-deparse-state-set-attrs state state-backup)

	  ;; TODO Loop until the line is broken correctly, but don't let
	  ;; fill-column reach value that lets little or no space on the
	  ;; line. But only try once if the related custom variable(s) allow
	  ;; for line break between opening paren and the following character
	  ;; or closing paren and the preceding character.
	  (let ((fill-column (1- fill-column)))
	    (pgqa-dump-maybe-insertion arg state indent-arg context))
	  )

	(when parens
	  ;; The closing parenthesis should be on a separate line, like the
	  ;; opening one.
	  (when
	      (and multiline
		   ;; Row expression is not considered a multi-line operator,
		   ;; so it looks better if the ")" is stuck to it.
		   ;;
		   ;; TODO Verify the behavior when the last expression of the
		   ;; row is a multi-line operator.
		   (or (null arg-is-list) (= (length (oref arg args)) 1))

		   ;; Subquery in an operator should have both opening and
		   ;; closing parenthesis attached tightly to it, just like it
		   ;; works for subquery in the FROM list.
		   (null arg-is-query))
	    (pgqa-deparse-newline state (1+ indent)))
	  ;; Never space in front of ")".
	  (oset state next-space 0)
	  (pgqa-deparse-string state ")" indent))

	(when (and is-unary (oref node postfix))
	  ;; TODO This part appears above for binary and unary prefix
	  ;; operators. Move it into a new function (pgqa-deparse-op-string?)
	  (if (oref node gui)
	      (pgqa-dump-start-gui (oref node gui) state))

	  (pgqa-deparse-string state op indent)

	  (if (oref node gui)
	      (pgqa-dump-end-gui (oref node gui) state)))

	(setq i (1+ i))

	(setq arg-multiline-prev arg-multiline))
      )
    )
  (pgqa-dump-end node state))

(defmethod pgqa-dump ((node pgqa-target-entry) state indent context)
  "Turn target list entry into a string."

  (pgqa-dump-start node state)
  (pgqa-dump (oref node expr) state indent context)

  (when (slot-boundp node 'alias)
    (if pgqa-print-as
	(pgqa-deparse-string state "AS" indent))
    (pgqa-dump (oref node alias) state indent context))

  (pgqa-dump-end node state))

(defmethod pgqa-dump ((node pgqa-alias) state indent context)
  "Turn target list entry alias into a string."

  (pgqa-dump-start node state)
  (pgqa-deparse-string state (oref node name) indent)
  (pgqa-dump-end node state))

(defmethod pgqa-dump ((node pgqa-alias-arg) state indent context)
  "Turn an argument of a target list entry alias into a string."

  (pgqa-dump-start node state)
  (pgqa-deparse-string state (oref node var) indent)
  (if (slot-boundp node 'datatype)
      (pgqa-dump (oref node datatype) state indent context))
  (pgqa-dump-end node state))

(provide 'pgqa-dump)
