;; Copyright (C) 2016 Antonin Houska
;;
;; This file is part of PGQA.
;;
;; PGQA is free software: you can redistribute it and/or modify it under the
;; terms of the GNU General Public License as published by the Free Software
;; Foundation, either version 3 of the License, or (at your option) any later
;; version.

;; PGQA is distributed in the hope that it will be useful, but WITHOUT ANY
;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;; FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
;; details.

;; You should have received a copy of the GNU General Public License qalong
;; with PGQA. If not, see <http://www.gnu.org/licenses/>.

(require 'eieio)

(defclass pgqa-node ()
  (
   (region
    :initarg :region
    :documentation "Start and end position of a nonterminal."
    )

   (markers
    :initarg :markers
    :documentation "Start and end position in the form of a marker."
    )

   (insertion
    :initarg :insertion
    :initform nil
    :documentation "An instance of `pgqa-query-string-insertion' if assigned."
    )
   )

  "Class representing a generic node of SQL query tree.")

;; Special node to carry information on region and markers of an operator
;; string. As it's a subclass of pgqa-node, it can be processed by
;; pgqa-setup-node-gui.
(defclass pgqa-gui-node (pgqa-node)
  (
   ;; Currently 'operator or 'func-call. Both for the sake of highlighting.
   (parent-kind
    :initarg :parent-kind
    :documentation "The kind of the owning node"
    )
   )
  )

(defmethod pgqa-node-walk ((node pgqa-gui-node) walker context)
  (funcall walker node context))

;; gui can be either a single instance of pgqa-gui-node or a list of them.
(defun pgqa-walk-gui (node walker context)
  (let ((gui (oref node gui)))
    (if (listp gui)
	(pgqa-node-walk-list gui walker context)
      (funcall walker gui context)))
  )

(defclass pgqa-expr (pgqa-node)
  (
   (args :initarg :args)
   )
  "A node representing an operation on one or multiple arguments.")

;; Besides attaching the markers to nodes, add them to pgqa-query-markers, to
;; make cleanup easier.
;;
;; context is currently used to indicate that the regions have been
;; initialized during a dump, which can't easily avoid leading whitespace. So
;; if the value is non-nil, we should skip that whitespace when setting up the
;; marker.
(defun pgqa-setup-node-gui (node context)
  "Turn region(s) into a markers to the node."
  (let* ((reg-vec (oref node region))
	 (reg-start (elt reg-vec 0))
	 (reg-end (elt reg-vec 1))
	 (m-start (make-marker))
	 (m-end (make-marker))
	 (o))

    ;; If the region might start with a whitespace, skip it (the whitespace).
    (when context
      (goto-char reg-start)
      (while (and (looking-at "\\s-\\|\$")
		  (< reg-start (point-max)))
	(setq reg-start (1+ reg-start))
	(goto-char reg-start)))

    (set-marker m-start reg-start)
    (set-marker m-end reg-end)

    ;; The insertion types are such that the start and end markers always span
    ;; only the original region.
    (set-marker-insertion-type m-start t)
    (set-marker-insertion-type m-end nil)

    ;; Keep track of markers.
    (push m-start pgqa-query-markers)
    (push m-end pgqa-query-markers)

    (oset node markers (vector m-start m-end)))
  )

;; For performance reasons (see the Overlays section of Elisp documentation)
;; we assign the face as text property, although cleanup would be simpler if
;; we assigned the face via overlay.
(defun pgqa-set-node-face (node context)
  (if (eq (eieio-object-class node) 'pgqa-gui-node)
      (if (slot-boundp node 'markers)
	  (let* ((m (oref node markers))
		 (m-start (elt m 0))
		 (m-end (elt m 1))
		 (kind (oref node parent-kind))
		 (face))

	    (if (eq kind 'operator)
		(setq face 'pgqa-operator-face)
	      (if (eq kind 'func-call)
		  (setq face 'pgqa-func-call-face)
		(error
		 (format
		  "Unrecognized kind of gui node parent: %s"
		  kind))))
	    (put-text-property m-start m-end
			       'font-lock-face face))
	)
    )
  )

;; Remove the face added previously by pgqa-set-node-face.
(defun pgqa-reset-node-face (node context)
  (if (eq (eieio-object-class node) 'pgqa-gui-node)
      (if (slot-boundp node 'markers)
	  (let* ((m (oref node markers))
		 (m-start (elt m 0))
		 (m-end (elt m 1)))
	    (remove-text-properties m-start m-end
				    '(font-lock-face nil)))
	)
    )
  )

;; An utility to apply a function to all nodes of a tree.
;;
;; If the walker function changes the nodes, caller is responsible for having
;; taken a copy of the original node.
;;
;; Currently it seems more useful to process the sub-nodes before the actual
;; node.
(defmethod pgqa-node-walk ((node pgqa-node) walker context)
  "Run the walker function on sub-nodes and the node itself"

  ;; It seems safer to force each node to implement this function explicitly
  ;; than to process the node w/o sub-nodes here.
  (error (format "walker method not implemented for %s class"
		 (eieio-object-class-name node))))

(defun pgqa-node-walk-list (node-list walker context)
  "Run the node walker function on each item of a list."
  (dolist (node node-list)
    (pgqa-node-walk node walker context)))

;; CAUTION! Keep `pgqa-copy-common-query-slots' up-to-date if adding new
;; slots.
(defclass pgqa-query (pgqa-node)
  (
   (kind :initarg :kind)
   (target-expr :initarg :target-expr
		:initform nil)
   (from-expr :initarg :from-expr
	      :initform nil)

   ;; An instance of `group-having-sort-part' class. Only used at parse time,
   ;; then converted to group-by, having and order-by below.
   (group-having-sort-part :initarg :group-having-sort-part
			   :initform nil)

   ;; The parts `group-having-sort-part'gets flattened into.
   (group-by :initarg :group-by
	     :initform nil)
   (having :initarg :having
	   :initform nil)
   (order-by :initarg :order-by
	     :initform nil)

   ;; A list used temporarily when parsing an INSERT query. The first element
   ;; is table name (eventually to be assigned to target-table' slot). If the
   ;; list has 2nd element, it's a list of columns (eventually to be assigned
   ;; to insert-cols slot).
   (insert-table-expr :initarg :insert-table-expr
		      :initform nil)

   ;; Table subject to INSERT / UPDATE / DELETE.
   (target-table :initarg :target-table :initform nil)

   ;; Column list if the query is INSERT command.
   (insert-cols :initarg :insert-cols :initform nil)

   ;; Either SELECT query or VALUES clause of INSERT command.
   (insert-input :initarg :insert-input :initform nil)
   )
  "A generic SQL query (or subquery).")

(defun pgqa-copy-common-query-slots (dst src)
  (oset dst target-expr (oref src target-expr))
  (if (oref src from-expr)
      (oset dst from-expr (oref src from-expr)))
  (if (oref src group-by)
      (oset dst group-by (oref src group-by)))
  (if (oref src having)
      (oset dst having (oref src having)))
  (if (oref src order-by)
      (oset dst order-by (oref src order-by))))

(defmethod pgqa-node-walk ((node pgqa-query) walker context)
  (if (oref node target-expr)
      (pgqa-node-walk (oref node target-expr) walker context))

  (if (oref node target-table)
      (pgqa-node-walk (oref node target-table) walker context))

  (if (oref node insert-input)
      (pgqa-node-walk (oref node insert-input) walker context))

  (if (oref node from-expr)
      (let ((fe (oref node from-expr)))
	(if (oref fe from-list)
	    (pgqa-node-walk-list (oref fe from-list) walker context))
	(if (slot-boundp fe 'qual)
	    (pgqa-node-walk (oref fe qual) walker context))))

  (if (oref node group-by)
    (pgqa-node-walk (oref node group-by) walker context))
  (if (oref node having)
    (pgqa-node-walk (oref node having) walker context))
  (if (oref node order-by)
    (pgqa-node-walk (oref node order-by) walker context))

  (funcall walker node context))

(defclass pgqa-from-expr (pgqa-node)
  (
   (from-list :initarg :from-list)
   (qual :initarg :qual)
   )
  "FROM expression of an SQL query."
)

;; A single argument represents table, function, subquery or VALUES clause. If
;; the 'args slot has elements, the FROM list entry is a join.
(defclass pgqa-from-list-entry (pgqa-expr)
  (
   ;; Instance of pgqa-from-list-entry-alias.
   (alias :initarg :alias)
   ;; For a simple entry, the value is one of "table", "function", "query",
   ;; "values". For join it's "left", "right", "full" (nil implies inner join
   ;; as long as 'args has 2 elements).
   (kind :initarg :kind)

   ;; Join expression if the entry is a join.
   (qual :initarg :qual)
   )
  "From list entry (table, join, subquery, ...)"
)

(defmethod pgqa-node-walk ((node pgqa-from-list-entry) walker context)
  (if (slot-boundp node 'alias)
      (funcall walker (oref node alias) context))

  ;; If node is a join, recurse into the sides and process qualifier.
  (let ((args (oref node args)))
    (if (= (length args) 2)
	(progn
	  (pgqa-node-walk (car args) walker context)
	  (pgqa-node-walk (car (cdr args)) walker context)
	  (pgqa-node-walk (oref node qual) walker context))
      ;; Process the single argument (query, function, ...).
      (cl-assert (= (length args) 1) nil
		 "from-list entry has incorrect number of arguments")
      (pgqa-node-walk (car args) walker context))
    (funcall walker node context)))

(defclass pgqa-alias (pgqa-node)
  (
   (name :initarg :name)
   )
  "Generic alias."
  )

(defclass pgqa-alias-arg (pgqa-node)
  (
   (var :initarg :var)

   ;; An instance of `pgqa-data-type'.
   (datatype :initarg :datatype)
   )
  "Alias output column."
  )

(defclass pgqa-from-list-entry-alias (pgqa-node)
  (
   ;; Instance of `pgqa-alias'.
   (name :initarg :name)

   ;; If :cols bound, it's an instance of `pgqa-node-list' of `pgqa-alias-arg'
   ;; items.
   (cols :initarg :cols)
   )
  "From list entry alias."
  )

;; This class exists primarily for pgq-dump. Top-level clauses that contain a
;; single expression should be represented by its subclass.
(defclass pgqa-top-clause (pgqa-node)
  (
   (expr :initarg :expr)
   )
  "Common class for some other which are dumped in similar way.")

(defmethod pgqa-node-walk ((node pgqa-top-clause) walker context)
  (pgqa-node-walk (oref node expr) walker context))

(defclass pgqa-group-clause (pgqa-top-clause)
  (
   )
  "GROUP BY clause.")

;; XXX Shouldn't `pgqa-group-clause' be just renamed to be more generic and
;; used here as well?
(defclass pgqa-having-clause (pgqa-top-clause)
  (
   )
  "HAVING clause.")

(defclass pgqa-sort-clause (pgqa-top-clause)
  (
   )
  "ORDER BY clause.")

;; Wrapper for GROUP BY and HAVING. Since we have a single rule for both (in
;; order to keep the number of rules for the query reasonably low), it makes
;; sense to have both clauses in a single structure. (List could do the work
;; but class makes tree walker(s) easier to read.)
;;
;; This class is only used at parse time. Once parsing is complete, it gets
;; flattened.
;;
;; As the contained clauses do have their GUI nodes, the only reason for this
;; class to inherit from `pgqa-node' is that it has to pass region info to the
;; containing query in the generic way.
(defclass pgqa-group-having-part (pgqa-node)
  (
   ;; Instance of `pgqa-group-clause'.
   (group :initarg :group :initform nil)
   ;; Instance of `pgqa-having-clause'.
   (having :initarg :having :initform nil))
  )

;; Like above, possibly including ORDER BY.
(defclass pgqa-group-having-sort-part (pgqa-node)
  (
   ;; Instance of `pgqa-group-having-part'.
   (group-having :initarg :group-having :initform nil)

   ;; Instance of `pgqa-sort-clause'.
   (sort :initarg :sort :initform nil))

  "GROUP BY (possibly including HAVING) and ORDER BY expressions.")

;; Flatten the parse-time classes like `pgqa-group-having-sort-part'.
(defun pgqa-flatten-query-tree (node context)
  (if (eq (eieio-object-class node) 'pgqa-query)
      (let ((ghs (oref node group-having-sort-part))
	    (gh)
	    (kind (oref node kind)))
	(when ghs
	  (setq gh (oref ghs group-having))
	  (when gh
	    (oset node group-by (oref gh group))
	    (oset node having (oref gh having)))
	  (oset node order-by (oref ghs sort))

	  ;; The slot is no longer needed.
	  (oset node group-having-sort-part nil))

	(when (string= kind "INSERT")
	  (let ((input (oref node insert-input))
		(ite (oref node insert-table-expr)))

	    ;; Copy the actual slots.
	    (pgqa-copy-common-query-slots node input)

	    (oset node target-table (car ite))
	    (if (cdr ite)
		(oset node insert-cols (nth 1 ite))))
	  )
	)
    )
  )

(defclass pgqa-func-call (pgqa-node)
  (
   (name :initarg :name)
   ;; arguments are stored as a single pgqa-operator having :op=",".
   (args :initarg :args)

   ;; Like pgqa-operator. XXX Is it worth to define a common parent class for
   ;; pgqa-func-call, pgqa-operator and possibly some other classes that might
   ;; be highlighted in the future?
   (gui :initarg :gui
	:initform nil)
   )
  "Function call"
)

(defmethod pgqa-node-walk ((node pgqa-func-call) walker context)
  (funcall walker (oref node name) context)
  (if (oref node args)
      (pgqa-node-walk (oref node args) walker context))
  (if (oref node gui)
    (pgqa-walk-gui node walker context))
  (funcall walker node context))

;; String constant.
(defclass pgqa-string (pgqa-node)
  (
   (value :initarg :value)
   )
  "A String.")

(defmethod pgqa-node-walk ((node pgqa-string) walker context)
  (funcall walker node context))

;; Number is currently stored as a string - should this be changed?
(defclass pgqa-number (pgqa-node)
  (
   (value :initarg :value)
   )
  "A number.")

(defmethod pgqa-node-walk ((node pgqa-number) walker context)
  (funcall walker node context))

(defclass pgqa-obj (pgqa-expr)
  (
   ;; The :args slot (inherited from pgqa-expr) contains the dot-separated
   ;; components of table / column reference.
   ;;
   ;; XXX Can't we simply use pgqa-expr class here?
   ;;
   ;; x.y expression can represent either column "y" of table "x" or table "y"
   ;; of schema "x". Instead of teaching parser to recognize the context (is
   ;; it possible?) let's postpone resolution till analysis phase.
   ;;
   ;; Note that the number of arguments is not checked during "raw parsing",
   ;; and that asterisk can be at any position, not only the last one.
   )
  "Table or column reference.")

(defmethod pgqa-node-walk ((node pgqa-obj) walker context)
  ;; The individual args are strings, so only process the alias.
  (funcall walker node context))

(defclass pgqa-data-type (pgqa-expr)
  (
   ;; The :args slot (inherited from pgqa-expr) contains a list of type name
   ;; components, each of which should be an instance of pgqa-obj. In most
   ;; cases it's a single-item list, but can have 2 items at least in one
   ;; special case ("double precision").
   )
  "Data type name.")

(defmethod pgqa-node-walk ((node pgqa-data-type) walker context)
  (funcall walker node context))

(defclass pgqa-case (pgqa-node)
  (
   (arg
    :initarg :arg
    :documentation "The expression following the CASE keyword.")

   (branches
    :initarg :branches
    :documentation "A list of `pgqa-case-branch' instances.")

   (else
    :initarg :else
    :documentation "The ELSE expression.")
   )

   "CASE expression")

(defmethod pgqa-node-walk ((node pgqa-case) walker context)
  (pgqa-node-walk-list (oref node branches) walker context)
  (funcall walker node context))

(defclass pgqa-case-branch (pgqa-expr)
  (
   ;; The (inherited) :args slot contains a 2-element list. The first and
   ;; second element represent arguments of the WHEN and THEN clause
   ;; respectively.
   )

  "The \"WHEN ... THEN ...\" part of CASE expression.")

(defmethod pgqa-node-walk ((node pgqa-case-branch) walker context)
  (pgqa-node-walk-list (oref node args) walker context)
  (funcall walker node context))

(defclass pgqa-operator (pgqa-expr)
  (
   ;; TODO Rename to 'code ?
   (op :initarg :op)

   ;; Region info and marker of the operator string(s) is stored separate so
   ;; that access to the string remains straightforward. Can be a single
   ;; instance of pgqa-gui-node or list of those.
   (gui :initarg :gui
	:initform nil)

   (prec :initarg :prec
	 :documentation "Operator precedence, for the sake of printing.")
   (postfix :initarg :postfix
	    :initform nil
	    :documentation "If the expression has only one argument, it's
considered to be an unary operator. This slot tells whether it's a postfix
operator. nil indicates it's a prefix operator.")
   )
  "Generic operator.")

(defmethod pgqa-node-walk ((node pgqa-operator) walker context)
  (pgqa-node-walk-list (oref node args) walker context)

  (if (oref node gui)
    (pgqa-walk-gui node walker context))
  (funcall walker node context))

;; Sublink is mostly treated like an operator, but it seems to deserve
;; separate class. The :op slot indicates the sublink kind (e.g. EXISTS), and
;; subquery is the only argument.
;;
;; TODO Simplify instance creation so that the subquery can be passed directly
;; instead of being wrapped in a singl-element list.
(defclass pgqa-sublink (pgqa-operator)
  (
   )
  "Sublink node")

;; Some nodes are actually plain lists, but it's easier to work with a class
;; because it has a name.
(defclass pgqa-node-list (pgqa-operator)
  (
   (op :initarg op :initform ",")
   (prec :initarg prec :initform pgqa-precedence-comma)
   )

  "List of nodes."
  )

(defclass pgqa-target-list (pgqa-node-list)
  (
   )
  "Query target list.")

(defclass pgqa-target-entry (pgqa-node)
  (
   (expr :initarg :expr)
   (alias :initarg :alias)
   )
  "Target list entry"
  )

(defmethod pgqa-node-walk ((node pgqa-target-entry) walker context)
  (pgqa-node-walk (oref node expr) walker context)
  (funcall walker node context))

(defclass pgqa-query-string-insertion ()
  (
   ;; The inserted string is used as a key. XXX There can be multiple
   ;; insertions with the same key. Should this slot be renamed or should the
   ;; insertions be organized in a different way than a single list (e.g. a
   ;; hash containing a list for each key)?
   (key :initarg :key)
   ;; Where in the buffer the insertion starts and ends.
   (start :initarg :start)
   (end :initarg :end)
   (overlay :initarg :overlay :initform nil))

  "If the input query has the form of SQL string, it can consist of parts
concatenated using the SQL concatenation operator (||). Each part is either an
SQL string or an insertion. For example, if PL/pgSQL language is used to
construct the query string, then PL/pgSQL variable representing part of the
string is the insertion.")

(defclass pgqa-string-to-query-context ()
  (
   (last-insertion :initarg :last-insertion
		   :initform nil)
   )

  "Input / output for `pgqa-dump-maybe-insertion'")

(provide 'pgqa-node)
